from libcpp.vector cimport vector
from libcpp cimport bool

cdef class LinkageModel:
    def __dealloc__(self):
        del self.c_inst 

cdef class Univariate(LinkageModel):
    def __cinit__(self):
        self.c_inst = new linkage_config_t()

cdef class BlockMarginalProduct(LinkageModel):
    def __cinit__(self,
        block_size : int = -1
    ):
        self.c_inst = new linkage_config_t(True,block_size)

cdef class Full(LinkageModel):
    def __cinit__(self):
        self.c_inst = new linkage_config_t(True,0)

cdef class StaticLinkageTree(LinkageModel):
    def __cinit__(self,
        maximum_set_size : int = -1
    ):
        cdef bool is_static = True 
        cdef int similarity_index = 2
        cdef bool filtered = True
        self.c_inst = new linkage_config_t(similarity_index, filtered, maximum_set_size, is_static)

cdef class LinkageTree(LinkageModel):
    def __cinit__(self,
        similarity_measure : string = b'MI',
        filtered : bool = False,
        maximum_set_size : int = -1
    ):
        cdef bool is_static = False
        cdef int similarity_index = -1
        if( similarity_measure == b'MI' ):
            similarity_index = 0
        elif( similarity_measure == b'NMI' ):
            similarity_index = 1
        else:
            raise AssertionError("Unknown similarity measure "+similarity_measure)
        self.c_inst = new linkage_config_t(similarity_index, filtered, maximum_set_size, is_static)

#cdef class Conditional(LinkageModel):
#    def __cinit__(self,
#        max_clique_size : int = 1,
#        cliques_as_fos_elements : bool = True,
#        include_full_fos_element : bool = True
#    ):
#        if not cliques_as_fos_elements and not include_full_fos_element:
#            raise AssertionError("At least one of input parameters 'cliques_as_fos_elements' or 'include_full_fos_element' must be True.")
#        self.c_inst = new linkage_config_t(max_clique_size,cliques_as_fos_elements,include_full_fos_element)

cdef class Custom(LinkageModel):
    def __cinit__(self,
        *args,
        **kwargs
    ):
        cdef string file = str.encode("")
        cdef vector[vector[int]] fos = []
        if 'file' in kwargs:
            file = str.encode(kwargs['file'])
        if 'fos' in kwargs:
            fos = kwargs['fos']
        if(len(file) is 0 and len(fos) is 0):
            raise AssertionError("Constructor requires exactly 1 argument.")
        if(len(file) > 0 and len(fos) > 0):
            raise AssertionError("Constructor requires exactly 1 argument.")
        if( len(file) > 0 ):
            self.c_inst = new linkage_config_t(file)
        elif( len(fos) > 0 ):
            self.c_inst = new linkage_config_t(fos)
        else:
            raise AssertionError("Constructor requires exactly 1 argument.")

#def UCondGG():
#    return Conditional(1,False,True)

#def UCondFG():
#    return Conditional(1,True,False)

#def UCondHG():
#    return Conditional(1,True,True)

#def MCondHG(*args,**kwargs):
#    if len(args) == 1:
#        return Conditional(args[0],True,True)
#    elif "max_clique_size" in kwargs:
#        return Conditional(kwargs["max_clique_size"],True,True)
#    else:
#        raise RuntimeError("Invalid arguments.")




import gomea
import numpy as np
from functools import cache

# Custom fitness function resembling the concatenated deceptive trap function of size k
class MaxCutGBO(gomea.fitness.GBOFitnessFunctionDiscrete):
    # Any members must be assigned in __new__ to make them accessible during instantiation of superclass
    def __new__(self, input_file, vtr_file ):
        number_of_variables = 0
        value_to_reach = 1e308
        self.edges = []
        with open(input_file,"r") as f:
            lines = [line.split() for line in f.readlines()]
            number_of_variables = int(lines[0][0])
            number_of_edges = int(lines[0][1])
            for i in range(number_of_edges):
                v1 = int(lines[i+1][0])-1
                v2 = int(lines[i+1][1])-1
                w = float(lines[i+1][2])
                self.edges.append((v1,v2,w))
            assert(number_of_edges == len(self.edges))
        with open(vtr_file,"r") as f:
            lines = f.readlines()
            value_to_reach = float(lines[0].strip())
        return super().__new__(self,number_of_variables,value_to_reach)

    @cache
    def number_of_subfunctions( self ):
        return len(self.edges)
    
    @cache
    def inputs_to_subfunction( self, subfunction_index ):
        return list(self.edges[subfunction_index][0:2])

    def subfunction(self, subfunction_index, variables):
        v1,v2,w = self.edges[subfunction_index]
        if variables[v1] == variables[v2]:
            return 0.0
        else:
            return w

class MaxCutBBO(gomea.fitness.BBOFitnessFunctionDiscrete):
    def __new__(self, input_file, vtr_file ):
        number_of_variables = 0
        value_to_reach = 1e308
        self.edges = []
        with open(input_file,"r") as f:
            lines = [line.split() for line in f.readlines()]
            number_of_variables = int(lines[0][0])
            number_of_edges = int(lines[0][1])
            for i in range(number_of_edges):
                v1 = int(lines[i+1][0])-1
                v2 = int(lines[i+1][1])-1
                w = float(lines[i+1][2])
                self.edges.append((v1,v2,w))
            assert(number_of_edges == len(self.edges))
        with open(vtr_file,"r") as f:
            lines = f.readlines()
            value_to_reach = float(lines[0].strip())
        return super().__new__(self,number_of_variables,value_to_reach)

    def objective_function(self, objective_index, variables):
        cut = 0.0
        for v1,v2,w in self.edges:
            if variables[v1] != variables[v2]:
                cut += w
        return cut


import gomea
import gc
import numpy as np
from output import readResultsFromJson, writeResultsToJson
import pathos.multiprocessing as mp 

from deceptive_trap.PythonTrapFunction import DeceptiveTrapFunctionGBO as PythonTrapFunctionGBO
from deceptive_trap.PythonTrapFunction import DeceptiveTrapFunctionBBO as PythonTrapFunctionBBO
from gomea.fitness import DeceptiveTrapFunction as CppTrapFunctionGBO
from gomea.fitness import DeceptiveTrapFunctionBBO as CppTrapFunctionBBO

from maxcut.PythonMaxCut import MaxCutGBO as PythonMaxCutGBO
from maxcut.PythonMaxCut import MaxCutBBO as PythonMaxCutBBO
from gomea.fitness import MaxCut as CppMaxCutGBO
from gomea.fitness import MaxCutBBO as CppMaxCutBBO

from soreb_chain.PythonSOREBChainStrong import SOREBChainStrongGBO as PythonSOREBChainStrongGBO
from soreb_chain.PythonSOREBChainStrong import SOREBChainStrongBBO as PythonSOREBChainStrongBBO
from gomea.fitness import SOREBChainStrong as CppSOREBChainStrongGBO
from gomea.fitness import SOREBChainStrongBBO as CppSOREBChainStrongBBO

from rosenbrock.PythonRosenbrock import RosenbrockFunctionGBO as PythonRosenbrockFunctionGBO
from rosenbrock.PythonRosenbrock import RosenbrockFunctionBBO as PythonRosenbrockFunctionBBO
from gomea.fitness import RosenbrockFunction as CppRosenbrockFunctionGBO
from gomea.fitness import RosenbrockFunctionBBO as CppRosenbrockFunctionBBO

def getFitnessFunction(prob_tag,lang_tag,dim=-1,input_file="",vtr_file=""):
    if prob_tag == 'trap5':
        if lang_tag == 'py':
            return PythonTrapFunctionGBO(dim,5)
        elif lang_tag == 'py_bbo':
            return PythonTrapFunctionBBO(dim,5)
        elif lang_tag == 'cpp':
            return CppTrapFunctionGBO(dim,5)
        elif lang_tag == 'cpp_bbo':
            return CppTrapFunctionBBO(dim,5)
    elif prob_tag == 'maxcut':
        if lang_tag == 'py':
            return PythonMaxCutGBO(input_file,vtr_file)
        elif lang_tag == 'py_bbo':
            return PythonMaxCutBBO(input_file,vtr_file)
        elif lang_tag == 'cpp':
            return CppMaxCutGBO(input_file,vtr_file)
        elif lang_tag == 'cpp_bbo':
            return CppMaxCutBBO(input_file,vtr_file)
    elif prob_tag == 'rosenbrock':
        if lang_tag == 'py':
            return PythonRosenbrockFunctionGBO(dim,1e-10)
        elif lang_tag == 'py_bbo':
            return PythonRosenbrockFunctionBBO(dim,1e-10)
        elif lang_tag == 'cpp':
            return CppRosenbrockFunctionGBO(dim,1e-10)
        elif lang_tag == 'cpp_bbo':
            return CppRosenbrockFunctionBBO(dim,1e-10)
    elif prob_tag == 'sorebchainstrong':
        if lang_tag == 'py':
            return PythonSOREBChainStrongGBO(dim,1e-10)
        elif lang_tag == 'py_bbo':
            return PythonSOREBChainStrongBBO(dim,1e-10)
        elif lang_tag == 'cpp':
            return CppSOREBChainStrongGBO(dim,1e-10)
        elif lang_tag == 'cpp_bbo':
            return CppSOREBChainStrongBBO(dim,1e-10)
    return None

def getLinkageModel(lm_tag):
    if lm_tag == 'bm5':
        return gomea.linkage.BlockMarginalProduct(block_size=5)
    elif lm_tag == 'uni':
        return gomea.linkage.Univariate()
    elif lm_tag == 'full':
        return gomea.linkage.Full()
    elif lm_tag == 'ucondhg':
        return gomea.linkage.UCondHG()
    elif lm_tag == 'lt':
        return gomea.linkage.LinkageTree()
    elif lm_tag == 'lt10':
        return gomea.linkage.LinkageTree(maximum_set_size=10)
    elif lm_tag == 'lt50':
        return gomea.linkage.LinkageTree(maximum_set_size=50)
    elif lm_tag == 'lt100':
        return gomea.linkage.LinkageTree(maximum_set_size=100)
    elif lm_tag == 'slt':
        return gomea.linkage.StaticLinkageTree()
    elif lm_tag == 'slt10':
        return gomea.linkage.StaticLinkageTree(maximum_set_size=10)
    elif lm_tag == 'slt50':
        return gomea.linkage.StaticLinkageTree(maximum_set_size=50)
    elif lm_tag == 'slt100':
        return gomea.linkage.StaticLinkageTree(maximum_set_size=100)

def getDomain(prob_tag):
    if prob_tag == 'trap5':
        return 'd'
    elif prob_tag == 'maxcut':
        return 'd'
    elif prob_tag == 'rosenbrock':
        return 'rv'
    elif prob_tag == 'sorebchainstrong':
        return 'rv'
    return None

def runGOMEA(prob_tag,lang_tag,lm_tag,dim,max_evals,max_sec,input_file,vtr_file):
    ff = getFitnessFunction(prob_tag,lang_tag,dim,input_file,vtr_file)
    lm = getLinkageModel(lm_tag)
    result = None
    if getDomain(prob_tag) == 'd':
        dgom = gomea.DiscreteGOMEA(fitness=ff,linkage_model=lm,max_number_of_evaluations=max_evals,max_number_of_seconds=max_sec)
        result = dgom.run()
        del dgom
    elif getDomain(prob_tag) == 'rv':
        rvgom = gomea.RealValuedGOMEA(fitness=ff,linkage_model=lm,max_number_of_evaluations=max_evals,max_number_of_seconds=max_sec)
        result = rvgom.run()
        del rvgom
    del ff
    del lm
    gc.collect()
    return result

def runExperiment(prob_tag,lang_tag,lm_tag,max_dim=1000,max_sec=-1,rerun=False,nthreads=1,nruns=10,input_file="",vtr_file=""):
    dim = 10
    cur_results = readResultsFromJson(prob_tag,lang_tag,lm_tag)
    fev_budget = 10000000
    fev_results = {}
    time_results = {}
    eval_time_results = {}
    bestf_results = {}
    succ_rate_results = {}
    if not rerun and 'fev' in cur_results.keys():
        fev_results = cur_results['fev']
    if not rerun and 't' in cur_results.keys():
        time_results = cur_results['t']
    if not rerun and 'bestf' in cur_results.keys():
        bestf_results = cur_results['bestf']
    if not rerun and 'eval_t' in cur_results.keys():
        eval_time_results = cur_results['eval_t']
    if not rerun and 'succ_rate' in cur_results.keys():
        succ_rate_results = cur_results['succ_rate']
    if len(input_file) > 0:
        dim = int(input_file[-14:-7])
    while dim <= max_dim or max_dim == -1:
        if rerun or (dim not in fev_results.keys()) or (dim not in time_results.keys()) or (dim not in eval_time_results.keys()) or (dim not in bestf_results.keys()) or (dim not in succ_rate_results.keys()):
            nsucc = 0
            fev_results[dim] = []
            time_results[dim] = []
            eval_time_results[dim] = []
            bestf_results[dim] = []
            succ_rate_results[dim] = 0

            thr_pool = mp.Pool(nthreads)
            tup = (prob_tag,lang_tag,lm_tag,dim,fev_budget,max_sec,input_file,vtr_file)
            args = ((tup,) * nruns )
            print(dim,prob_tag,lang_tag,lm_tag)
            results = thr_pool.starmap(runGOMEA,args)
            nsucc = 0
            vtr = getFitnessFunction(prob_tag,lang_tag,dim,input_file,vtr_file).value_to_reach
            domain = getDomain(prob_tag)
            for res in results:
                fev_results[dim].append(res['evaluations'][-1])
                time_results[dim].append(res['time'][-1])
                eval_time_results[dim].append(res['eval_time'][-1])
                bestf_results[dim].append(res['best_obj_val'][-1])
                if (domain == 'd' and res['best_obj_val'][-1] >= vtr) or (domain == 'rv' and res['best_obj_val'][-1] <= vtr):
                    nsucc += 1
            succ_rate_results[dim] = nsucc / nruns
            results = {"t" : time_results, "fev" : fev_results, "bestf" : bestf_results, "eval_t" : eval_time_results, "succ_rate" : succ_rate_results }
            writeResultsToJson(results,prob_tag,lang_tag,lm_tag)
            print("Dim",dim," ",nsucc/nruns," suc ",np.median(time_results[dim]),"sec",np.median(fev_results[dim]),"fev")
        dim = 2*dim
        if len(input_file) > 0:
            break



# distutils: language = c++
# cython: c_string_type=unicode, c_string_encoding=utf8

from gomea.output cimport OutputStatisticsWrapper
from gomea.output import OutputStatistics
from gomea.discrete cimport gomeaIMS, Config
from gomea.linkage cimport LinkageModel, StaticLinkageTree
from libcpp.string cimport string
from libcpp cimport bool
from cpython.exc cimport PyErr_CheckSignals
from tqdm import tqdm
import inspect

from gomea.fitness cimport FitnessFunction, GBOFitnessFunction, BBOFitnessFunction
include "EmbeddedFitness.pxi"

# Create a Cython extension type which holds a C++ instance
# as an attribute and create a bunch of forwarding methods
# Python extension type.
cdef class DiscreteGOMEA:
    cdef gomeaIMS c_inst  # Hold a C++ instance which we're wrapping
    cdef Config c_config
    cdef FitnessFunction fitness

    def __cinit__(self,
        # Optimization problem settings (required)
        fitness: FitnessFunction, 
        # GOMEA settings
        linkage_model : LinkageModel = StaticLinkageTree(),
        max_number_of_populations : int = 25,
        IMS_subgeneration_factor: int = 4,
        base_population_size: int = 2,
        max_number_of_evaluations : int = -1,
        max_number_of_generations : int = -1,
        max_number_of_seconds : float = -1.0,
        random_seed : int = -1,
    ):

        # Initialize attributes 
        #_, _, _, values = inspect.getargvalues(inspect.currentframe())
        #for arg, val in values.items():
            #setattr(self, arg, val)

        self.c_config = Config()
        assert( (<FitnessFunction?>fitness).c_inst_discrete != NULL, "FitnessFunction is not discrete." )
        self.c_config.fitness = (<FitnessFunction?>fitness).c_inst_discrete
        # as we only store a pointer to a field of the FitnessFunction (which actually manages the memory around the pointer)
        # there is the risk of the FitnessFunction being removed & its contents deallocated. This leaves the above pointer dangling.
        # This is particularly likely if the FitnessFunction was constructed as an argument to this class -
        # in which case it & corresponding memory will be deallocated after this initialization procedure has completed.
        # Fix this by keeping the reference to the FitnessFunction around for as long as this instance exists.
        self.fitness = fitness

        self.c_config.linkage_config = linkage_model.c_inst
        self.c_config.folder = string(b"output_discrete_gomea")
        self.c_config.maximumNumberOfGOMEAs = max_number_of_populations
        self.c_config.IMSsubgenerationFactor = IMS_subgeneration_factor
        self.c_config.basePopulationSize = base_population_size
        self.c_config.problemInstancePath = string(b"") 
        self.c_config.maxArchiveSize = 1000 
        self.c_config.maximumNumberOfEvaluations = max_number_of_evaluations
        self.c_config.maximumNumberOfGenerations = max_number_of_generations
        self.c_config.maximumNumberOfSeconds = max_number_of_seconds
        self.c_config.AnalyzeFOS = 0
        #if analyze_fos:
        #    self.c_config.AnalyzeFOS = 1
        #self.c_config.verbose = False 
        self.c_config.fix_seed = False
        if random_seed != -1:
            self.c_config.randomSeed = random_seed
            self.c_config.fix_seed = True

        # Initialize C++ instance
        self.c_inst = gomeaIMS(&self.c_config)

    def get_progress(self):
        return self.c_inst.getProgressUntilTermination()

    def check_termination(self):
        cdef bool t = self.c_inst.checkTermination()
        return t

    def run_generation(self):
        self.c_inst.runGeneration()

    def init_progress_bar(self):
        progress_bar = tqdm(desc="DiscreteGOMEA",unit='%',total=100)
        return progress_bar

    def update_progress_bar(self, progress_bar):
        progress_bar.n = self.get_progress()
        progress_bar.refresh()
    
    def run(self):
        self.c_inst.run()
        return OutputStatistics(OutputStatisticsWrapper.from_ptr(&self.c_inst.output))

    def run_with_progress(self):
        with self.init_progress_bar() as progress_bar:
            while( not self.check_termination() ):
                self.run_generation()
                self.update_progress_bar(progress_bar)
                #print("PyErr:",PyErr_CheckSignals() )
                if( PyErr_CheckSignals() == -1 ):
                    break
            self.update_progress_bar(progress_bar)
        #self.c_inst.run()



from experiments import runExperiment
from output import readResultsFromJson
from plotting import plotEvalTimeScalability, plotTimePerFEV, plotTimeScalability, plotFEVScalability

def runExperimentsTrap5BlockMarginal(max_dim=1000,rerun=False,nthreads=1,nruns=10,max_sec=-1):
    runExperiment('trap5','cy','bm5',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    #runExperiment('trap5','cyc','bm5',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','cpp','bm5',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','cpp_bbo','bm5',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','py','bm5',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','py_bbo','bm5',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)

def runExperimentsTrap5LT(max_dim=1000,rerun=False,nthreads=1,nruns=10,max_sec=-1):
    runExperiment('trap5','cy','lt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    #runExperiment('trap5','cyc','lt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','cpp','lt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','cpp_bbo','lt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','py','lt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','py_bbo','lt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)

def runExperimentsTrap5StaticLT(max_dim=1000,rerun=False,nthreads=1,nruns=10,max_sec=-1):
    runExperiment('trap5','py','slt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','cy','slt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    #runExperiment('trap5','cyc','slt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','cpp','slt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)

def runExperiments():
    rerun_experiments = False 
    nthreads = 15
    nruns = 30
    max_sec = 3600
    #runExperimentsTrap5LT(max_dim=5000,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns)
    #runExperimentsTrap5BlockMarginal(max_dim=10000,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns)
    #runExperimentsTrap5StaticLT(max_dim=5000,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns)
    #runExperimentsTrap5LT(max_dim=10000,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    #runExperimentsTrap5BlockMarginal(max_dim=10000,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    #runExperimentsTrap5StaticLT(max_dim=10000,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','py','slt',max_dim=10000,rerun=False,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','py','lt',max_dim=10000,rerun=False,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','py','lt50',max_dim=10000,rerun=False,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','py_bbo','lt',max_dim=2000,rerun=False,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','cpp','slt',max_dim=10000,rerun=False,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','cpp','lt50',max_dim=5000,rerun=False,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','cpp','lt',max_dim=10000,rerun=False,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('trap5','cpp_bbo','lt',max_dim=5000,rerun=False,nthreads=nthreads,nruns=nruns,max_sec=max_sec)

def makePlots(out_dir="plots/"):
    res_py_bm5 = readResultsFromJson('trap5','py','bm5')
    res_cy_bm5 = readResultsFromJson('trap5','cy','bm5')
    #res_cyc = readResultsFromJson('trap5','cyc','bm5')
    res_cpp_bm5 = readResultsFromJson('trap5','cpp','bm5')
    res_pybbo_bm5 = readResultsFromJson('trap5','py_bbo','bm5')
    res_cppbbo_bm5 = readResultsFromJson('trap5','cpp_bbo','bm5')
    #results = [("GBO-Python",res_py_bm5),("GBO-C++",res_cpp_bm5),("BBO-Python",res_pybbo_bm5),("BBO-C++",res_cppbbo_bm5)]
    #xticks = [10*2**i for i in range(20)]
    xticks = []
    #plotTimePerFEV(results,out_dir=out_dir,tag='trap5_bm5',max_dim=10000,xticks=xticks)
    #plotEvalTimeScalability(results,out_dir=out_dir,tag='trap5_bm5',max_dim=10000,xticks=xticks)
    #plotTimeScalability(results,out_dir=out_dir,tag='trap5_bm5',max_dim=10000,xticks=xticks)
    #plotFEVScalability(results,out_dir=out_dir,tag='trap5_bm5',max_dim=10000,xticks=xticks)

    res_py_lt = readResultsFromJson('trap5','py','lt')
    res_py_lt100 = readResultsFromJson('trap5','py','lt100')
    res_cy_lt100 = readResultsFromJson('trap5','cy','lt100')
    #res_cyc = readResultsFromJson('trap5','cyc','lt')
    res_cpp_lt = readResultsFromJson('trap5','cpp','lt')
    res_cpp_lt50 = readResultsFromJson('trap5','cpp','lt50')
    res_cpp_lt100 = readResultsFromJson('trap5','cpp','lt100')
    res_pybbo_lt100 = readResultsFromJson('trap5','py_bbo','lt100')
    res_cppbbo_lt = readResultsFromJson('trap5','cpp_bbo','lt')
    res_cppbbo_lt50 = readResultsFromJson('trap5','cpp_bbo','lt50')
    res_cppbbo_lt100 = readResultsFromJson('trap5','cpp_bbo','lt100')
    #results = [("Python",res_py_lt100),("C++",res_cpp_lt100),("Python(BBO)",res_pybbo_lt100),("C++(BBO)",res_cppbbo_lt100)]
    #plotTimePerFEV(results,out_dir=out_dir,tag='trap5_lt100',max_dim=10000,xticks=xticks)
    #plotTimeScalability(results,out_dir=out_dir,tag='trap5_lt100',max_dim=10000,xticks=xticks)
    #plotEvalTimeScalability(results,out_dir=out_dir,tag='trap5_lt100',max_dim=10000,xticks=xticks)
    #plotFEVScalability(results,out_dir=out_dir,tag='trap5_lt100',max_dim=10000,xticks=xticks)
    
    res_py_slt = readResultsFromJson('trap5','py','slt')
    #res_cy_slt = readResultsFromJson('trap5','cy','slt')
    res_cpp_slt = readResultsFromJson('trap5','cpp','slt')
    res_pybbo_lt = readResultsFromJson('trap5','py_bbo','lt')
    res_cppbbo_lt = readResultsFromJson('trap5','cpp_bbo','lt')
    results = [("GBO-Python",res_py_lt),("GBO-C++",res_cpp_lt),("BBO-Python",res_pybbo_lt),("BBO-C++",res_cppbbo_lt)]
    #plotTimePerFEV(results,out_dir=out_dir,tag='trap5_slt',max_dim=10000,xticks=xticks)
    plotEvalTimeScalability(results,out_dir=out_dir,tag='trap5_lt',max_dim=10000,xticks=xticks)
    #plotTimeScalability(results,out_dir=out_dir,tag='trap5_slt',max_dim=10000,xticks=xticks)
    #plotFEVScalability(results,out_dir=out_dir,tag='trap5_slt',max_dim=10000,xticks=xticks)
    
    results = [("GBO-LT",res_cpp_lt),("GBO-SLT",res_cpp_slt),("BBO-LT",res_cppbbo_lt)]
    #results = [("BBO-LT",res_cppbbo_lt),("BBO-LT(50)",res_cppbbo_lt50),("BBO-LT(100)",res_cppbbo_lt100),("GBO-SLT",res_cpp_slt),("GBO-LT",res_cpp_lt),("GBO-LT(50)",res_cpp_lt50),("GBO-LT(100)",res_cpp_lt100)]
    #plotEvalTimeScalability(results,out_dir=out_dir,tag='trap5_cpp_lms',max_dim=10000,xticks=xticks)
    #plotTimeScalability(results,out_dir=out_dir,tag='trap5_cpp_lms',max_dim=10000,xticks=xticks)
    plotFEVScalability(results,out_dir=out_dir,tag='trap5_cpp_lms',max_dim=10000,xticks=xticks)
    
    results = [("GBO-LT",res_py_lt),("GBO-SLT",res_py_slt),("BBO-LT",res_pybbo_lt)]
    #results = [("BBO-LT",res_cppbbo_lt),("BBO-LT(50)",res_cppbbo_lt50),("BBO-LT(100)",res_cppbbo_lt100),("GBO-SLT",res_cpp_slt),("GBO-LT",res_cpp_lt),("GBO-LT(50)",res_cpp_lt50),("GBO-LT(100)",res_cpp_lt100)]
    #plotEvalTimeScalability(results,out_dir=out_dir,tag='trap5_py_lms',max_dim=10000,xticks=xticks)
    plotTimeScalability(results,out_dir=out_dir,tag='trap5_py_lms',max_dim=10000,xticks=xticks)
    #plotFEVScalability(results,out_dir=out_dir,tag='trap5_py_lms',max_dim=10000,xticks=xticks)

if __name__ == '__main__':
    #runExperiments()
    makePlots("plots/trap5/")


import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.ticker import ScalarFormatter

matplotlib.rc('font', size=12)
matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['ps.fonttype'] = 42

def addResultToPlot(results,succ_rate_results,label):
    dimensions = np.sort(np.int_(list(results.keys())))
    dimensions = [d for d in dimensions if d in succ_rate_results and succ_rate_results[d] > 0.99]
    medians = [np.median(results[dim]) for dim in dimensions]
    err_high = [np.percentile(results[dim],90)-np.median(results[dim]) for dim in dimensions] 
    err_low = [np.median(results[dim])-np.percentile(results[dim],10) for dim in dimensions] 
    plt.errorbar(dimensions,medians,yerr=[err_low,err_high],capsize=3,label=label)

def addTimePerEvalResultToPlot(time_results,fev_results,succ_rate_results,label):
    dimensions = np.sort(list(time_results.keys()))
    dimensions = [d for d in dimensions if d in succ_rate_results and succ_rate_results[d] > 0.99]
    tpf_results = {dim:np.divide(time_results[dim],fev_results[dim]) for dim in dimensions}
    medians = [np.median(tpf_results[dim]) for dim in dimensions]
    err_high = [np.percentile(tpf_results[dim],90)-np.median(tpf_results[dim]) for dim in dimensions] 
    err_low = [np.median(tpf_results[dim])-np.percentile(tpf_results[dim],10) for dim in dimensions] 
    plt.errorbar(dimensions,medians,yerr=[err_low,err_high],capsize=3,label=label)

def plotFEVScalability(results,out_dir="plots/",tag='',min_dim=8,max_dim=1000,yrange=[],xticks=[]):
    fig,ax = plt.subplots(dpi=150)
    for label,res in results:
        addResultToPlot(res['fev'],res['succ_rate'],label=label)
    fig.patch.set_facecolor('white')
    plt.yscale('log')
    plt.xscale('log',base=10)
    plt.xlabel("Dimensionality")
    plt.ylabel("Number of evaluations")
    if len(xticks)>0:
        plt.xticks(xticks)
    plt.ylim(5,1e7)
    if len(yrange)>0:
        plt.ylim(*yrange)
    plt.xlim(min_dim,max_dim)
    plt.grid(linewidth=0.5)
    formatter = plt.ScalarFormatter()
    formatter.set_scientific(False)
    ax.xaxis.set_major_formatter(formatter)
    ax.legend(loc='lower right')
    plt.savefig(out_dir+'/fev_'+tag+'.pdf',bbox_inches='tight')

def plotEvalTimeScalability(results,out_dir="plots/",tag='',min_dim=8,max_dim=1000,yrange=[],xticks=[]):
    fig,ax = plt.subplots(dpi=150)
    for label,res in results:
        addResultToPlot(res['eval_t'],res['succ_rate'],label=label)
    fig.patch.set_facecolor('white')
    plt.yscale('log')
    plt.xscale('log',base=10)
    plt.xlabel("Dimensionality")
    plt.ylabel("Evaluation time (s)")
    if len(xticks)>0:
        plt.xticks(xticks)
    if len(yrange)>0:
        plt.ylim(*yrange)
    #plt.ylim(0.0005,30)
    plt.xlim(min_dim,max_dim)
    plt.grid(linewidth=0.5)
    formatter = plt.ScalarFormatter()
    formatter.set_scientific(False)
    ax.xaxis.set_major_formatter(formatter)
    ax.legend(loc='lower right')
    plt.savefig(out_dir+'/evalt_'+tag+'.pdf',bbox_inches='tight')

def plotTimeScalability(results,out_dir="plots/",tag='',min_dim=8,max_dim=1000,yrange=[],xticks=[]):
    fig,ax = plt.subplots(dpi=150)
    for label,res in results:
        addResultToPlot(res['t'],res['succ_rate'],label=label)
    fig.patch.set_facecolor('white')
    plt.yscale('log')
    plt.xscale('log',base=10)
    plt.xlabel("Dimensionality")
    plt.ylabel("Time (s)")
    if len(xticks)>0:
        plt.xticks(xticks)
    if len(yrange)>0:
        plt.ylim(*yrange)
    #plt.ylim(0.0005,30)
    plt.xlim(min_dim,max_dim)
    plt.grid(linewidth=0.5)
    formatter = plt.ScalarFormatter()
    formatter.set_scientific(False)
    ax.xaxis.set_major_formatter(formatter)
    ax.legend(loc='lower right')
    plt.savefig(out_dir+'/t_'+tag+'.pdf',bbox_inches='tight')

def plotTimePerFEV(results,out_dir="plots/",tag='',min_dim=8,max_dim=1000,xticks=[],yrange=[]):
    fig,ax = plt.subplots(dpi=150)
    for label,res in results:
        #print(label,res)
        addTimePerEvalResultToPlot(res['t'],res['fev'],res['succ_rate'],label=label)
    fig.patch.set_facecolor('white')
    plt.yscale('log')
    plt.xscale('log',base=10)
    plt.xlabel("Dimensionality")
    plt.ylabel("Time (s) / Evaluations ")
    if len(xticks)>0:
        plt.xticks(xticks)
    if len(yrange)>0:
        plt.ylim(*yrange)
    #plt.ylim(0.0005,30)
    plt.xlim(min_dim,max_dim)
    plt.grid(linewidth=0.5)
    formatter = plt.ScalarFormatter()
    formatter.set_scientific(False)
    ax.xaxis.set_major_formatter(formatter)
    ax.legend(loc='lower right')
    plt.savefig(out_dir+'/tperfev_'+tag+'.pdf',bbox_inches='tight')



import numpy as np
from setuptools import setup, Extension
from Cython.Build import cythonize

extensions = []
extensions.append( Extension("CythonMaxCut",
        ["CythonMaxCut.pyx"],
        language="c++")
)

setup(
    ext_modules = cythonize(extensions,
        include_path = ["."] + [np.get_include()],
        annotate = True,
        language_level = "3"),
    zip_safe = False
)



import numpy as np
cimport numpy as np
from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp cimport bool

cdef extern from "gomea/src/common/linkage_config.hpp" namespace "gomea":
    cdef cppclass linkage_config_t:
        linkage_config_t() except +
        linkage_config_t(int,bool) except +
        linkage_config_t(int,bool,int,bool) except +
        linkage_config_t(int,bool,bool) except +
        linkage_config_t(vector[vector[int]]) except +
        linkage_config_t(string) except +

cdef class LinkageModel:
    cdef linkage_config_t *c_inst

cdef class Univariate(LinkageModel):
    pass

cdef class BlockMarginalProduct(LinkageModel):
    pass

cdef class Full(LinkageModel):
    pass

cdef class LinkageTree(LinkageModel):
    pass

cdef class StaticLinkageTree(LinkageModel):
    pass

#cdef class Conditional(LinkageModel):
#    pass

cdef class Custom(LinkageModel):
    pass


from experiments import runExperiment
from output import readResultsFromJson
from plotting import plotEvalTimeScalability, plotTimePerFEV, plotTimeScalability, plotFEVScalability

def runExperimentsRosenbrockUni(max_dim=1000,rerun=False,nthreads=1,nruns=10,max_sec=-1):
    runExperiment('rosenbrock','cpp','uni',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    #runExperiment('rosenbrock','cpp_bbo','uni',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    #runExperiment('rosenbrock','py','uni',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    #runExperiment('rosenbrock','py_bbo','uni',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)

def runExperimentsRosenbrockLT(max_dim=1000,rerun=False,nthreads=1,nruns=10,max_sec=-1):
    runExperiment('rosenbrock','cpp','lt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('rosenbrock','cpp_bbo','lt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('rosenbrock','py','lt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('rosenbrock','py_bbo','lt',max_dim=160,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('rosenbrock','py_bbo','lt50',max_dim=160,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('rosenbrock','cpp_bbo','lt50',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)

def runExperimentsRosenbrockSLT(max_dim=1000,rerun=False,nthreads=1,nruns=10,max_sec=-1):
    #runExperiment('rosenbrock','py','slt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    #runExperiment('rosenbrock','cpp','slt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    #runExperiment('rosenbrock','cpp','slt50',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('rosenbrock','py','slt50',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)

def runExperiments():
    rerun_experiments = False 
    nthreads = 15
    nruns = 30
    max_sec = 3600
    runExperimentsRosenbrockUni(max_dim=1000,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperimentsRosenbrockLT(max_dim=400,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperimentsRosenbrockSLT(max_dim=1000,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns,max_sec=max_sec)

def makePlots(out_dir="plots/"):
    xticks = []
    yrange = [1e3,1e8]
    #res_py_lt = readResultsFromJson('rosenbrock','py','lt')
    #res_cpp_lt = readResultsFromJson('rosenbrock','cpp','lt')
    #res_cppbbo_lt = readResultsFromJson('rosenbrock','cpp_bbo','lt')
    
    #res_py_slt = readResultsFromJson('rosenbrock','py','slt')
    res_py_slt50 = readResultsFromJson('rosenbrock','py','slt50')
    res_cpp_slt50 = readResultsFromJson('rosenbrock','cpp','slt50')
    #res_cy_slt = readResultsFromJson('rosenbrock','cy','slt')
    #res_cpp_slt = readResultsFromJson('rosenbrock','cpp','slt')
    #res_pybbo_lt = readResultsFromJson('rosenbrock','py_bbo','lt')
    #res_cppbbo_lt = readResultsFromJson('rosenbrock','cpp_bbo','lt')
    res_cppbbo_lt50 = readResultsFromJson('rosenbrock','cpp_bbo','lt50')
    res_pybbo_lt50 = readResultsFromJson('rosenbrock','py_bbo','lt50')
    res_cppbbo_uni = readResultsFromJson('rosenbrock','cpp_bbo','uni')
    res_cpp_uni = readResultsFromJson('rosenbrock','cpp','uni')
    res_py_uni = readResultsFromJson('rosenbrock','py','uni')
    res_pybbo_uni = readResultsFromJson('rosenbrock','py_bbo','uni')
    results = [("GBO-Python",res_py_uni),("GBO-C++",res_cpp_uni),("BBO-Python",res_pybbo_uni),("BBO-C++",res_cppbbo_uni)]
    #plotTimePerFEV(results,out_dir=out_dir,tag='rosenbrock_uni',max_dim=1000,xticks=xticks)
    plotEvalTimeScalability(results,out_dir=out_dir,tag='rosenbrock_uni',max_dim=1000,xticks=xticks)
    #plotTimeScalability(results,out_dir=out_dir,tag='rosenbrock_uni',max_dim=1000,xticks=xticks)
    #plotFEVScalability(results,out_dir=out_dir,tag='rosenbrock_uni',max_dim=1000,xticks=xticks,yrange=yrange)

    results = [("GBO-Uni",res_cpp_uni),("GBO-SLT(50)",res_cpp_slt50),("BBO-Uni",res_cppbbo_uni),("BBO-LT(50)",res_cppbbo_lt50)]
    #results = [("BBO-LT",res_cppbbo_lt),("BBO-LT(50)",res_cppbbo_lt50),("BBO-LT(100)",res_cppbbo_lt100),("GBO-SLT",res_cpp_slt),("GBO-LT",res_cpp_lt),("GBO-LT(50)",res_cpp_lt50),("GBO-LT(100)",res_cpp_lt100)]
    #plotEvalTimeScalability(results,out_dir=out_dir,tag='rosenbrock_cpp_lms',max_dim=1000,xticks=xticks)
    #plotTimeScalability(results,out_dir=out_dir,tag='rosenbrock_cpp_lms',max_dim=1000,xticks=xticks)
    plotFEVScalability(results,out_dir=out_dir,tag='rosenbrock_cpp_lms',max_dim=1000,xticks=xticks,yrange=yrange)
    
    #results = [("BBO-LT",res_pybbo_lt),("GBO-SLT",res_py_slt),("GBO-LT",res_py_lt)]
    results = [("GBO-Uni",res_py_uni),("GBO-SLT(50)",res_py_slt50),("BBO-Uni",res_pybbo_uni),("BBO-LT(50)",res_pybbo_lt50)]
    #results = [("BBO-LT",res_pybbo_lt),("GBO-LT",res_py_lt)]
    #results = [("BBO-LT",res_cppbbo_lt),("BBO-LT(50)",res_cppbbo_lt50),("BBO-LT(100)",res_cppbbo_lt100),("GBO-SLT",res_cpp_slt),("GBO-LT",res_cpp_lt),("GBO-LT(50)",res_cpp_lt50),("GBO-LT(100)",res_cpp_lt100)]
    #plotEvalTimeScalability(results,out_dir=out_dir,tag='rosenbrock_py_lms',max_dim=1000,xticks=xticks)
    plotTimeScalability(results,out_dir=out_dir,tag='rosenbrock_py_lms',max_dim=1000,xticks=xticks)
    #plotFEVScalability(results,out_dir=out_dir,tag='rosenbrock_py_lms',max_dim=1000,xticks=xticks,yrange=yrange)

if __name__ == '__main__':
    #runExperiments()
    makePlots("plots/rosenbrock/")


from gomea.output cimport output_statistics_t
from ctypes import * 

cdef class OutputStatisticsWrapper:
    def __cinit__(self):
        pass

    def __init__(self):
        raise TypeError("This class cannot be instantiated directly.")

    @staticmethod
    cdef OutputStatisticsWrapper from_ptr(output_statistics_t *_ptr):
        cdef OutputStatisticsWrapper wrapper = OutputStatisticsWrapper.__new__(OutputStatisticsWrapper)
        wrapper.c_ptr = _ptr
        return wrapper

class OutputStatistics:
    def __init__(self, stats : OutputStatisticsWrapper):
        self.metrics = stats.c_ptr[0].getAllMetricNames()
        self.metrics_dict = {}

        def getMetricValues( metric_name ):
            try:
                return stats.c_ptr[0].getMetricValues[int](metric_name)
            except Exception as err:
                pass
            try:
                return stats.c_ptr[0].getMetricValues[double](metric_name)
            except Exception as err:
                pass
            try:
                b_result = stats.c_ptr[0].getMetricValues[string](metric_name)
                return [byte_string.decode('utf-8') for byte_string in b_result]
            except Exception as err:
                pass

        for metric_name in self.metrics:
            s_metric_name = metric_name.decode()
            self.metrics_dict[s_metric_name] = getMetricValues(metric_name)
        self.metrics = [s.decode() for s in self.metrics]
        #self.elitist_genotype = stats.c_ptr[0].getElitistGenotype()
        #self.elitist_fitness = stats.c_ptr[0].getElitistFitness()

    def __getitem__(self, metric):
        assert( metric in self.metrics, str(metric)+" is not a valid metric." )
        return self.metrics_dict[metric]

    def getFinalStatistics(self):
        final_stats = {}
        for metric in self.metrics:
            final_stats[metric] = self.metrics_dict[metric][-1]
        return final_stats
    
    def printFinalStatistics(self):
        final_stats = self.getFinalStatistics()
        print("Final statistics:")
        for k,v in final_stats.items():
            print(k,":",v)
    
    def printAllStatistics(self):
        print("All statistics:")
        for k,v in self.metrics_dict.items():
            print(k,":",v)



# distutils: language=c++
# cython: language_level=3, boundscheck=False, wraparound=False

import cython
from libcpp.vector cimport vector
import numpy as np
cimport numpy as np
from gomea.fitness cimport GBOFitnessFunction
from gomea.fitness cimport pyGBOFitnessFunction_t
from cpython cimport PyObject

cdef class DeceptiveTrapFunction(GBOFitnessFunction):
    # Any members must be assigned in __new__ to make them accessible during instantiation of superclass
    cdef int k
    def __cinit__(self, number_of_variables : cython.int, k : cython.int ):
        assert( number_of_variables % k == 0 )
        self.k = k
        self.number_of_variables = number_of_variables
        self.value_to_reach = number_of_variables
        self.c_inst_discrete = new pyGBOFitnessFunction_t[char](number_of_variables,self.value_to_reach,<PyObject*>self)

    cpdef int number_of_subfunctions( self ):
        return int(self.number_of_variables / self.k)
    
    cpdef vector[int] inputs_to_subfunction( self, int subfunction_index ):
        cdef vector[int] indices = vector[int](self.k)
        cdef int i
        for i in range(0,self.k,1):
            indices[i] = self.k*subfunction_index + i
        return indices

    cpdef double subfunction(self, int subfunction_index, np.ndarray variables ):
        cdef vector[int] trap_variables = self.inputs_to_subfunction(subfunction_index)
        cdef int unitation = 0
        for v in trap_variables:
            unitation = unitation + variables[v]
        if unitation == self.k:
            return unitation
        else:
            return self.k - unitation - 1




import numpy as np
from setuptools import setup, Extension
from Cython.Build import cythonize
from pathlib import Path
import os
import gomea

extensions = []

libs_path = str(Path(gomea.__path__[0]).parent.absolute())
gomea_path = str(Path(gomea.__path__[0]).absolute())
libs_to_find = ['fitness', 'discrete', 'linkage', 'output']
libs_to_link = []
for root, dirs, files in os.walk(gomea_path):
    for lib_to_find in libs_to_find:
        for f in files:
            if f.startswith(lib_to_find) and f.endswith('.so'):
                libs_to_link.append("-l:"+str(f))
#print(libs_to_link)
# find names of library files within gomea_path to link with and add to 'libs_to_link'
extensions.append( Extension("CythonTrapFunction",
        ["CythonTrapFunction.pyx"],
        language="c++")
)
extensions.append( Extension("CythonTrapFunctionCDEF",
        ["CythonTrapFunction-cdef.pyx"],
        include_dirs=[libs_path],
        extra_compile_args=libs_to_link,
        extra_link_args=libs_to_link,
        library_dirs=[gomea_path],
        language="c++")
)

setup(
    ext_modules = cythonize(extensions,
        include_path = ["."] + [libs_path] + [np.get_include()],
        annotate = True,
        language_level = "3"),
    zip_safe = False
)



import gomea
import numpy as np

# Custom fitness function resembling the concatenated deceptive trap function of size k
class CustomTrapFunction(gomea.fitness.GBOFitnessFunctionDiscrete):
    # Any members must be assigned in __new__ to make them accessible during instantiation of superclass
    def __new__(self, number_of_variables, k, value_to_reach):
        assert( number_of_variables % k == 0 )
        self.k = k
        return super().__new__(self,number_of_variables,value_to_reach)

    def number_of_subfunctions( self ):
        return self.number_of_variables // self.k
    
    def inputs_to_subfunction( self, subfunction_index ):
        return range(self.k*subfunction_index,self.k*subfunction_index+self.k)

    def subfunction(self, subfunction_index, variables):
        trap_variables = variables[self.inputs_to_subfunction(subfunction_index)]
        unitation = np.sum(trap_variables)
        if unitation == self.k:
            return unitation
        else:
            return self.k - unitation - 1

dim = 20
lm = gomea.linkage.StaticLinkageTree(maximum_set_size=5)
fd = CustomTrapFunction(dim,k=5,value_to_reach=dim)
dgom = gomea.DiscreteGOMEA(fitness=fd,linkage_model=lm,max_number_of_evaluations=100000)
result = dgom.run()
result.printAllStatistics()
result.printFinalStatistics()



import gomea
import numpy as np

# Custom BBO fitness function resembling the Rosenbrock function
class CustomRosenbrockFunction(gomea.fitness.BBOFitnessFunctionRealValued):
    def objective_function( self, objective_index, variables ):
        f = 0
        for i in range(len(variables)-1):
            x = variables[i]
            y = variables[i+1]
            f += 100*(y-x*x)*(y-x*x) + (1.0-x)*(1.0-x)
        return f

dim = 10
frv = CustomRosenbrockFunction(dim,value_to_reach=1e-6)
lm = gomea.linkage.Univariate()
rvgom = gomea.RealValuedGOMEA(fitness=frv, linkage_model=lm, lower_init_range=-115, upper_init_range=-100, max_number_of_populations=1, base_population_size=100, max_number_of_evaluations=1000000)
result = rvgom.run()
result.printAllStatistics()
result.printFinalStatistics()



from experiments import runExperiment
from output import readResultsFromJson
from plotting import plotEvalTimeScalability, plotTimePerFEV, plotTimeScalability, plotFEVScalability

def runExperimentsSOREBChainFull(max_dim=1000,rerun=False,nthreads=1,nruns=10,max_sec=-1):
    runExperiment('sorebchainstrong','cpp','full',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('sorebchainstrong','cpp_bbo','full',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('sorebchainstrong','py','full',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('sorebchainstrong','py_bbo','full',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)

def runExperimentsSOREBChainLT(max_dim=1000,rerun=False,nthreads=1,nruns=10,max_sec=-1):
    runExperiment('sorebchainstrong','cpp','lt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('sorebchainstrong','cpp_bbo','lt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('sorebchainstrong','py','lt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('sorebchainstrong','py_bbo','lt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('sorebchainstrong','py_bbo','lt50',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('sorebchainstrong','cpp_bbo','lt50',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)

def runExperimentsSOREBChainSLT(max_dim=1000,rerun=False,nthreads=1,nruns=10,max_sec=-1):
    #runExperiment('sorebchainstrong','py','slt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('sorebchainstrong','cpp','slt',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)

def runExperimentsSOREBChainUCond(max_dim=1000,rerun=False,nthreads=1,nruns=10,max_sec=-1):
    runExperiment('sorebchainstrong','cpp','ucondhg',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperiment('sorebchainstrong','py','ucondhg',max_dim=max_dim,rerun=rerun,nthreads=nthreads,nruns=nruns,max_sec=max_sec)

def runExperiments():
    rerun_experiments = False 
    nthreads = 15
    nruns = 30
    max_sec = 3600
    runExperimentsSOREBChainSLT(max_dim=20,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperimentsSOREBChainUCond(max_dim=40,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperimentsSOREBChainFull(max_dim=40,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns,max_sec=max_sec)
    runExperimentsSOREBChainLT(max_dim=20,rerun=rerun_experiments,nthreads=nthreads,nruns=nruns,max_sec=max_sec)

def makePlots(out_dir="plots/"):
    res_py_lt = readResultsFromJson('sorebchainstrong','py','lt')
    res_py_lt100 = readResultsFromJson('sorebchainstrong','py','lt100')
    res_cpp_lt = readResultsFromJson('sorebchainstrong','cpp','lt')
    res_cpp_lt50 = readResultsFromJson('sorebchainstrong','cpp','lt50')
    res_cpp_lt100 = readResultsFromJson('sorebchainstrong','cpp','lt100')
    res_pybbo_lt100 = readResultsFromJson('sorebchainstrong','py_bbo','lt100')
    res_cppbbo_lt = readResultsFromJson('sorebchainstrong','cpp_bbo','lt')
    res_cppbbo_lt50 = readResultsFromJson('sorebchainstrong','cpp_bbo','lt50')
    res_cppbbo_lt100 = readResultsFromJson('sorebchainstrong','cpp_bbo','lt100')
    results = [("Python",res_py_lt100),("C++",res_cpp_lt100),("Python(BBO)",res_pybbo_lt100),("C++(BBO)",res_cppbbo_lt100)]
    plotTimePerFEV(results,out_dir=out_dir,tag='sorebchainstrong_lt100',max_dim=10000,xticks=xticks)
    plotTimeScalability(results,out_dir=out_dir,tag='sorebchainstrong_lt100',max_dim=10000,xticks=xticks)
    plotEvalTimeScalability(results,out_dir=out_dir,tag='sorebchainstrong_lt100',max_dim=10000,xticks=xticks)
    plotFEVScalability(results,out_dir=out_dir,tag='sorebchainstrong_lt100',max_dim=10000,xticks=xticks)
    
    res_py_slt = readResultsFromJson('sorebchainstrong','py','slt')
    #res_cy_slt = readResultsFromJson('sorebchainstrong','cy','slt')
    res_cpp_slt = readResultsFromJson('sorebchainstrong','cpp','slt')
    res_pybbo_lt = readResultsFromJson('sorebchainstrong','py_bbo','lt')
    res_cppbbo_lt = readResultsFromJson('sorebchainstrong','cpp_bbo','lt')
    #results = [("Python",res_py_slt),("Cython",res_cy_slt),("C++",res_cpp_slt),("Python(BBO)",res_pybbo_lt),("C++(BBO)",res_cppbbo_lt)]
    #plotTimePerFEV(results,out_dir=out_dir,tag='sorebchainstrong_slt',max_dim=10000,xticks=xticks)
    #plotEvalTimeScalability(results,out_dir=out_dir,tag='sorebchainstrong_slt',max_dim=10000,xticks=xticks)
    #plotTimeScalability(results,out_dir=out_dir,tag='sorebchainstrong_slt',max_dim=10000,xticks=xticks)
    #plotFEVScalability(results,out_dir=out_dir,tag='sorebchainstrong_slt',max_dim=10000,xticks=xticks)

    xticks = [] 
    results = [("BBO-LT",res_cppbbo_lt),("GBO-SLT",res_cpp_slt),("GBO-LT",res_cpp_lt)]
    #results = [("BBO-LT",res_cppbbo_lt),("BBO-LT(50)",res_cppbbo_lt50),("BBO-LT(100)",res_cppbbo_lt100),("GBO-SLT",res_cpp_slt),("GBO-LT",res_cpp_lt),("GBO-LT(50)",res_cpp_lt50),("GBO-LT(100)",res_cpp_lt100)]
    plotEvalTimeScalability(results,out_dir=out_dir,tag='sorebchainstrong_cpp_lms',max_dim=10000,xticks=xticks)
    plotTimeScalability(results,out_dir=out_dir,tag='sorebchainstrong_cpp_lms',max_dim=10000,xticks=xticks)
    plotFEVScalability(results,out_dir=out_dir,tag='sorebchainstrong_cpp_lms',max_dim=10000,xticks=xticks)
    
    results = [("BBO-LT",res_pybbo_lt),("GBO-SLT",res_py_slt),("GBO-LT",res_py_lt)]
    #results = [("BBO-LT",res_cppbbo_lt),("BBO-LT(50)",res_cppbbo_lt50),("BBO-LT(100)",res_cppbbo_lt100),("GBO-SLT",res_cpp_slt),("GBO-LT",res_cpp_lt),("GBO-LT(50)",res_cpp_lt50),("GBO-LT(100)",res_cpp_lt100)]
    plotEvalTimeScalability(results,out_dir=out_dir,tag='sorebchainstrong_py_lms',max_dim=10000,xticks=xticks)
    plotTimeScalability(results,out_dir=out_dir,tag='sorebchainstrong_py_lms',max_dim=10000,xticks=xticks)
    plotFEVScalability(results,out_dir=out_dir,tag='sorebchainstrong_py_lms',max_dim=10000,xticks=xticks)

if __name__ == '__main__':
    runExperiments()
    makePlots("plots/sorebchainstrong/")